// Copyright 2010, 2011, 2013 by Jon Dart. All Rights Reserved.
#include "board.h"
#include "notation.h"
#include "legal.h"
#include "hash.h"
#include "globals.h"
#include "chessio.h"
#include "util.h"
#include "search.h"
#include "arasvec.h"
#include "eco.h"
#include <iostream>
#include <fstream>
#include <utility>
#include <ctype.h>
#include <map>

using namespace std;

static const int MARGIN = 70;
static int MAX_PLY = 24;
static int MIN_OCCURRENCE = 6;

int CDECL main(int argc, char **argv)
{
   Bitboard::init();
   initOptions(argv[0]);
   Attacks::init();
   Scoring::init();
   if (!initGlobals(argv[0], false)) {
      cleanupGlobals();
      exit(-1);
   }
   atexit(cleanupGlobals);

   options.book.book_enabled = options.log_enabled = 0;

   int arg = 1;
   unsigned minMoves = 30;
   int minELO = 0;

   map<uint64,int> *hashTable = new map<uint64,int>();
   ECO e;

   while (arg < argc && *argv[arg] == '-') {
       char c = argv[arg][1];
       switch(c) {
       case 'p':
           ++arg;
           MAX_PLY = atoi(argv[arg]);
           ++arg;
           break;
       case 'n': 
           ++arg;
           MIN_OCCURRENCE = atoi(argv[arg]);
           ++arg;
           break;
       default: 
          cerr << "Invalid switch: " << c << endl;
          break;
       }
   }
   if (arg >= argc) {
      cerr << "usage: testpos -p <ply> -n <minCount> <input file>" << endl;
      return -1;
   }
   else {
      for (;arg < argc;++arg) {
         if (argv[arg][0] == '-') {
         }
         else
            break;
      }
      SearchController *searcher = new SearchController();
      Statistics stats;
      int round = 1;
      for (;arg < argc;arg++) {
         cerr << "processing " << argv[arg] << endl;
         ifstream pgn_file( argv[arg], ios::in);
         int c;
         int count = 0;
         hash_t game_hash;
         ColorType side;
         string result, white, black;
         if (!pgn_file.good()) {
            cerr << "could not open file " << argv[arg] << endl;
            exit(-1);
         }
         else {
            ArasanVector<ChessIO::Header> hdrs;

            Board board;
            hash_t lastHash = (hash_t)0;
            int valid = 1;
            while (!pgn_file.eof()) {

               long first;
               board.reset();
               side = White;
               while (pgn_file.good() && (c = pgn_file.get()) != EOF) {
                  if (c=='[') {
                     pgn_file.putback(c);
                     break;
                  }
               }
               hdrs.removeAll();
               ChessIO::collect_headers(pgn_file,hdrs,first);

               game_hash = 0;
               int mate = 0;
               int var = 0;
               int ply = 0;
               MoveArray moves;
               for (;;) {
                  static char buf[2048];
                  static char num[20];
                  ChessIO::Token tok = ChessIO::get_next_token(pgn_file,buf,2048);
                  if (tok == ChessIO::Eof)
                     break;
                  if (tok == ChessIO::Number) {
                    if (var) continue;
                    strncpy(num,buf,20);
                  }
                  else if (tok == ChessIO::GameMove) {
                     if (var) continue; 
                     // parse the move
                     Move m;
                     m = Notation::value(board,board.sideToMove(),buf);
                     if (IsNull(m) ||
                         !legalMove(board,StartSquare(m), DestSquare(m))) {
                                                  // echo to both stdout and stderr
                        cerr << "Illegal move: " << buf << endl;
                        valid = 0;
                        exit(0);

                     }
                     else if (valid) {
                        string result;
                        Notation::image(board,m,result);
                        BoardState bs = board.state;
                        board.doMove(m);
                        if (++ply <= MAX_PLY) {
                           moves.add_move(board,bs,m,result,false);
                           lastHash = board.hashCode();
                        }
                     }
                     side = OppositeColor(side);
                  }
                  else if (tok == ChessIO::Unknown) {
                     if (strcmp(buf,"(") == 0)
                          ++var;
                     else if (strcmp(buf,")")==0) {
                          --var;   
                     } else {
                       cerr << "Unrecognized text: " << buf << endl;
                       valid = 0;
                     }
                  }
                  else if (tok == ChessIO::Comment) {
                  }
                  else if (tok == ChessIO::Result) {
                      break; // all done
                  }
               }
               // done with a game
               if (ply < MAX_PLY + 6 || !valid) {
                  continue;
               }
               map<uint64,int,std::less<hash_t>>::const_iterator it =
                   hashTable->find(lastHash);
               if (it == hashTable->end()) {
                   // not found, add entry with count 1
                   (*hashTable)[lastHash] = 1;
               } else {
                   (*hashTable)[lastHash]++;
               }
               int freq = (*hashTable)[lastHash];
               if (freq == MIN_OCCURRENCE) {
                   ArasanVector<ChessIO::Header> newHdrs;
                   newHdrs.append(ChessIO::Header("Event","?"));
                   newHdrs.append(ChessIO::Header("Site","?"));
                   newHdrs.append(ChessIO::Header("Date","????.??.??"));
                   stringstream roundstr;
                   roundstr << round;
                   newHdrs.append(ChessIO::Header("Round",roundstr.str().c_str()));
                   newHdrs.append(ChessIO::Header("Result","*"));
                   // output a game
                   string eco, name;
                   e.classify(moves,eco,name);
                   newHdrs.append(ChessIO::Header("ECO",eco.c_str()));
                   ChessIO::store_pgn(cout,moves,White,"*",newHdrs);
                   round++;
               }
            }
            pgn_file.close();
         }
      }
   }

   return 0;
}
